// CForth  project ************************************************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.h","cforth.c","cforth.h",
// "usb-fs-042.c" and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32f0xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{
uint32_t recv;
uint8_t sendByte;

    RCC->AHBENR |= RCC_AHBENR_GPIOAEN;  // Enable Port A clock
    RCC->AHBENR |= RCC_AHBENR_GPIOBEN;  // Enable Port B clock

    GPIOA->MODER = 0x00000000;      //primarily all input
    GPIOB->MODER = 0x00000000;      //primarily all input
    GPIOA->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown
    GPIOB->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown

#ifdef LQFP32
//pin config repeated specifically:
	GPIOA->PUPDR &= ~(GPIO_PUPDR_PUPDR8);	//pull-down off, wasts current
    GPIOA->MODER &= ~(GPIO_MODER_MODER8);
    GPIOA->MODER |= GPIO_MODER_MODER8_0 ;  	//set PortA8 (LED) as output
    GPIOA->BSRR =  GPIO_BSRR_BS_8;        	//PA8 (LED) ON
#else
//pin config repeated specifically:
#ifdef MIDI						//specifically means "MINI",but VERO model too
	GPIOA->PUPDR &= 0xC3FFFFFF;	//PA14/13 HiZ
	GPIOA->PUPDR |= 0x14000000;	//PA14/13 Pull-UP for mode check
#endif
	GPIOB->PUPDR &= ~(GPIO_PUPDR_PUPDR8);	//pull-down off, wasts current
    GPIOB->MODER &= ~(GPIO_MODER_MODER8);
    GPIOB->MODER |= GPIO_MODER_MODER8_0 ;  	//set PortB8 (LED) as output
    GPIOB->BSRR =  GPIO_BSRR_BS_8;        	//PB8 (LED) ON
// for USB, remap PA11,12 to PA9,10
    RCC->APB2ENR |= RCC_APB2ENR_SYSCFGCOMPEN;
    SYSCFG->CFGR1 |= SYSCFG_CFGR1_PA11_PA12_RMP;
#endif
    gSYSCOUNT = 50;
    gLEDCOUNT = 0;
    gDELAYCOUNT = 0;
    gTIMEOUT = 0;
    gVid = DEFAULT_VID;
    gPid = DEFAULT_PID;
    guDeviceClass = CDC_CLASS;
    gUsbChecked = 0;
	gProject = 0;
	gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = DEFAULT_NUMBASE;
//these are maintained as dummies, depends if RS_232 or not:
    gSplit = 0;
    gBaud = DEFAULT_BAUD;
#ifdef MIDI
	gTempBaud = gBaud;
	gChannel = 0;
	gMIDI_HEAD = 0;
    gMIDI_TAIL = 0;
	gTransparent = 0;		//type check in startUSB() easier then
#endif
	gfRandom = 314159265;		//seed for xorshift32 algorithm
    SysTick_Config(SystemCoreClock/1000);

    msDelay(10);

//start Independent Watchdog  **************
    IWDG->KR = 0x00005555;
    IWDG->RLR = 0x0000FFFF;
    msDelay(2);
    IWDG->PR = 0x00000002;      //resulting timeout ca. 1-2 seconds
    msDelay(2);
    IWDG->KR = 0x0000CCCC;      //start Independent Watchdog

//init I/O state  *************************
//    op_ABORT();			//reset default Forth system data
    popFlashProject(0);
				//here checked only Baudrate,NumBase,USB Vid/Pid, project no.
    msDelay(2);


 //init communication  ********************
#ifdef LQFP32
	initUsart1();
	msDelay(2);
	sendModel();				//exclusively via USART;
#else
  #ifdef MIDI
	gTransparent = ~(GPIOA->IDR >> 13) & 0x00000003;
//jumper level			inverted
//11=none	->			gTransparent = 00 =0: CForth
//10=PA14none,PA13low->	gTransparent = 01 =1: USB-RS232 interface
//01=PA14low,PA13none->	gTransparent = 10 =2: conventional USB-MIDI interface
//00=PA14low PA13low->	gTransparent = 11 =3: USB_CDC to USART-MIDI interface
	if (gTransparent == 2)  gSplit = 4;//conventional USB/MIDI interface
	if (gTransparent >= 2) gBaud = 31;	//overrides popFlashProject() Baudrate
					//conventional USB/MIDI interface or USB_CDC to USART MIDI
  #endif

	initUsart2();
	msDelay(2);

  #ifdef MIDI
	if (gTransparent == 0)  sendModel(); //exclusively via USART	
										//i.e. PA13 and PA14=High:CForth
  #else
	sendModel();				//exclusively via USART;
  #endif
#endif

    msDelay(2);
    guConfigured = 0;

#ifdef LQFP32
	if ((GPIOA->IDR & GPIO_IDR_9) && (gVid != 0) && (gPid != 0)) startUSB();
	else  {
		sendString(gNoUsbStr);
		sendTB(CRR);
	}
#else
    startUSB();
  #ifdef MIDI
	if ((gTransparent == 1) || (gTransparent == 3)) {		
									//1=COM transparent (PA13=High)
									//3=USB=COM,USART=MIDI (PA13/14=low)
		while(1)  {					//USB works in CDC_CLASS
			recv = pullUsartRxBuf();
			if (recv)  sendUsbByte((uint8_t)(recv & 0x000000FF));
			recv = pullUsbRxBuf();
			if (recv)   sendB((uint8_t)(recv & 0x000000FF));
		}
	}
	if (gTransparent == 2)  {		//2=MIDI transparent (PA14=Low,PA13=High)
									//USB works in MIDI_CLASS
		while(1)  {	
			recv = pullUsartRxBuf();
			if (recv)  enCodeMidi((uint8_t)(recv & 0x000000FF));
			recv = pullUsbRxBuf();
			if (recv)   sendB((uint8_t)(recv & 0x000000FF));
		}
	}
// if (gTransparent==0) -activate CForth
  #endif
#endif

    popFlashProject(1);			//esentially show op_QUEST

    while(1) {
		quit();
		query();
		if (gfTibMax == 0) continue;
		eval();
    }
}

//##########################################################################

// Misc Low Level Routines
//************************

void sendModel(void)  // ****************************************************
// **************************************************************************
{
const uint8_t typeStr [] = "CForth for STM32F042 v.";
const uint8_t copyStr [] = " (c)2020-24 W.Schemmert";

    sendTB(CRR);
    sendString(typeStr);
    sendTB(HARDWARE_VERSION + 0x30);
    sendTB('.');
    sendNS32(REVNUM);
    sendString(copyStr);
    sendTB(CRR);
}


void LED_Off(void)  // ******************************************************
// built as subroutine to supply service to "usb-fs-042.c" hardware independent
// **************************************************************************
{
#ifdef LQFP32
        GPIOA->BSRR = GPIO_BSRR_BR_8; // clear bit PA8,LED OFF
#else
        GPIOB->BSRR = GPIO_BSRR_BR_8; // clear bit PB8,LED OFF
#endif
        gLEDCOUNT = 4; //ca 300 milliseconds
}


void msDelay(uint32_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
    gDELAYCOUNT = count;
    while(gDELAYCOUNT)  servSystem();
}


void wait5us(void)   // *****************************************************
// blocking delay ca 5us
// *************************************************************************
{
uint32_t countdown;

	countdown = 51;
    while (countdown)   {
        IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
        countdown-- ;
    }
}


void servSystem(void)   // ************************************************
//not very time critical but permanently repeated functions for system control
//specially called in pullUsartRxBuf and pullUsbRxBuf
//this way it is regularily served when idle waiting for external input
//but watchdog resets when it hangs somewhere else
//additionally slowed down by SysTick to 50 times/second
// ****************************************************************************
{
static uint32_t sSecCount;
	if (gSYSCOUNT == 0)  {		//countdown in SysTick_Handler()
		gSYSCOUNT = 50;         //restart 20Hz service interval
//feed watchdog
		IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

//LED control
		if (gLEDCOUNT) gLEDCOUNT--;
#ifdef LQFP32
		if (gLEDCOUNT==0)  GPIOA->BSRR =  GPIO_BSRR_BS_8; //PA8 (LED) ON again
#else
		if (gLEDCOUNT==0)  GPIOB->BSRR =  GPIO_BSRR_BS_8; //PB8 (LED) ON again
#endif
#ifdef LQFP32
		if ((GPIOA->IDR & GPIO_IDR_9) && (gVid != 0) && (gPid != 0))  {
#endif
//USB mode control
		sSecCount++;
		if (sSecCount == 20)  {          //restart 1 second service interval
			sSecCount = 0;

			if (guConfigured == 0) {
#ifndef MIDI
				gSplit = 0;			//avoid possible deadlock
				if (gBaud == 31)  {	//re-init BAUD??
					gBaud = DEFAULT_BAUD;
  #ifdef LQFP32
					initUsart1();
  #else
					initUsart2();
  #endif
				}
#endif

				if (gUsbChecked == 0) {
					startUSB();	//USB plugged, enforce re-init USB
					sendTB(CRR);
				}
			}
#ifdef MIDI
			else  {
				if ((gSplit == 4) && (guDeviceClass == CDC_CLASS))  {
						stopUSB();
						guDeviceClass = MIDI_CLASS;
						msDelay(500);
						startUSB();
				}
				if ((gSplit != 4) && (guDeviceClass == MIDI_CLASS))  {
					stopUSB();
					guDeviceClass = CDC_CLASS;
					msDelay(500);
					startUSB();
				}
			}
#endif
		}
#ifdef LQFP32
		}
#endif
	}
}


void startUSB(void)  // ******************************************************
// **************************************************************************
{
uint32_t oldTimeout;
const uint8_t cdcClassStr [] = "USB as virtual COM=";
#ifdef MIDI
const uint8_t midiClassStr [] = "USB as MIDI interface=" ;
#endif

    gUsbChecked = 1;
#ifdef MIDI
    if (gSplit == 4) guDeviceClass = MIDI_CLASS;
	else guDeviceClass = CDC_CLASS;
#else
	guDeviceClass = CDC_CLASS;
#endif
    initUSB();
    gTIMEOUT = 3000;
	oldTimeout = gTIMEOUT;
	while (gTIMEOUT)  {
		IWDG->KR = 0x0000AAAA;		//feed watchdog, timeout ca.0.5c
		if (gTIMEOUT != oldTimeout)  {	//check once per ms
			oldTimeout = gTIMEOUT;
			if ((gTIMEOUT & 0x00000003FF) == 0x3FF)  sendTB('.');
		}
        if (guConfigured)  break;
    }

#ifdef MIDI
		if (gTransparent == 0)  {	//i.e. only if CForth active
			if (gSplit == 4)  sendString(midiClassStr);
			else sendString(cdcClassStr);
			if (guConfigured)  {
				sendTB('o');
				sendTB('k');
			}
			else  sendString(gErrorStr);
		}
#else
	sendString(cdcClassStr);
    if (guConfigured)  {
        sendTB('o');
        sendTB('k');
    }
    else  sendString(gErrorStr);
#endif
	sendTB(CRR);
    msDelay(200);
}


void popFlashProject(uint32_t mode)   //*************************************
// **************************************************************************
{
uint32_t k, n;
uint32_t  *flashPtr;
uint32_t *sramPtr;

    flashPtr = (uint32_t *)PAGE29;
//4 words=16bytes for system and peripherial parameters
    k = ~(*flashPtr);		//cannot be==0 if saved project in Flash
    flashPtr++;
    n =  ~(*flashPtr);		//contains gBaud and (gfSpiBits>>2) (value 1...8)
							//cannot be==0 if project saved in Flash

	flashPtr++;
	if ((k | n ) == 0) {
		emptyForth();		//load default empty project if virgin Flash
		if (mode == 0) return;
	}
	else  {
		gProject = 1;
		gVid = (k >> 16) & 0x0000FFFF;
		gPid = k & 0x0000FFFF;
		if (gVid == 0) gVid = DEFAULT_VID;
		if (gPid == 0) gPid = DEFAULT_PID;

		if (n & 0x00000001)  gfNumBase = 16;
		else gfNumBase = 10;
		gfTempNumBase = gfNumBase;
		gBaud = (n >> 1) & 0x000007F;
#ifdef MIDI
		gTempBaud = gBaud;
#endif

//up to here, basic system parameters are used fpr empty project, too
		if (mode == 0)  return;

// ---------------------------------------------------
//from here only project#1 loaded
		gfSpiMode = (n >> 9) & 0x00000003;	
		gfSpiBits = ((n >> 11) & 0x0000001F) + 4;	//range 4...32
		gfSpiClock = (n >> 16) & 0x0000FFFF;

		k = ~(*flashPtr);
		flashPtr++;
		gfAuto = (k & 0x0000FFFF);

//		sendTB('?');
//		send0X32(gfAuto);

#ifdef MIDI
		gChannel = (k >> 16) & 0x0000001F;
#else
		gfFreq = (k >> 16) & 0x0000FFFF;
#endif
		k = ~(*flashPtr);
		flashPtr++;
#ifndef MIDI
		n = k & 0x0000FFFF;
		if (n > 2048) n |= 0xFFFF0000;		//manual cast to int32_t
		gfPWM1 = (int32_t)n;

		if (gfFreq) {
			subFREQ();
			subPWM1();
  #ifndef MOTOR
			n = (k >> 16) & 0x0000FFFF;
			if (n > 2048) n |= 0xFFFF0000;		//manual cast to int32_t
			gfPWM2 = (int32_t)n;
			subPWM2();
  #endif
		}
#endif

//loading is possible over page limits.
//gfUserHeader[] and gfTcode[] occupy a coherent data space in flash
		sramPtr = (uint32_t *) &(gfUserHeader[0]); //252words=1008bytes + 16=1024
		for (k = 0;k < (MAXUSER * 4);k++)  {
			*sramPtr = ~(*flashPtr);
			sramPtr++;
			flashPtr++;
		}
		sramPtr = (uint32_t *) &(gfTCode[0]);
		for (k = 0;k < 0x180;k++)  {
			*sramPtr = ~(*flashPtr);
			sramPtr++;
			flashPtr++;
		}
	}

    op_ABORT();
    if (gfAuto) {
//emergency brake to avoid AUTO deadlock
#ifdef LQFP32
		if((GPIOA->IDR & 0x00000400)==0) gfAuto=0; //if PA10 (RX input)is low
									//bugfix 10Jan24
#else
		if((GPIOA->IDR & 8) == 0) gfAuto = 0;	//if PA3 (RX input)is low
#endif
    }

    op_QUEST();
    if (gfAuto) {
        gfParsed = gfAuto - 1;	//-1 to distinguish from empty AUTOEXE == 0
        interpretUserToken();
    }
}


void pushFlashProject(void)   //*********************************************
// **************************************************************************
{
uint32_t k;
uint16_t  *flashPtr;
uint32_t *sramPtr;

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB
    while ((FLASH->SR & FLASH_SR_BSY) !=0);//wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
            FLASH->KEYR = KEY1;
            FLASH->KEYR = KEY2;
    }

//erase Page29
    FLASH->CR |= FLASH_CR_PER;
    FLASH->AR = PAGE29;
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    FLASH->CR &= ~FLASH_CR_PER;

    FLASH->CR |= FLASH_CR_PG;
//write system parameters and gfUserHeader[] into page29
//programming Flash is possible only by 16bit half-words
//first programmed (even address) 16bit half-word will be read as lower 16bits
//                                          when data are read as 32bit words!!
//second programmed 16bit half-word will be read as higher 16bits
//                                          when data are read as 32bit words!!
//programming from here: word read, 2*halfWord burn

    flashPtr = (uint16_t *)PAGE29;
    //8 half-words=16bytes for system and peripheral parameters
    *flashPtr = (uint16_t)(~gPid & 0x0000FFFF);	//read as lower 16bits
//							gPid and gVid will be used for Flash empty check
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;

    *flashPtr = (uint16_t)(~gVid & 0x0000FFFF);	//read as upper 16bits
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
    flashPtr++;

	if (gfNumBase == 10) k = 0;
    else k = 1;
	k |= ((gBaud & 0x0000007F) << 1); // | ((gRsTerm & 1) << 8);
	k |= ((gfSpiMode & 0x0000003)<< 9) ; 
    k |= (((gfSpiBits - 4) & 0x0000001F) << 11); //burn range = 0...28
	*flashPtr = (uint16_t)(~k & 0x0000FFFF); //program first=read lower16bit
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	flashPtr++;

	k =  gfSpiClock;
	*flashPtr = (uint16_t)(~k & 0x0000FFFF); //program last=read upper16bit
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	flashPtr ++;	

    *flashPtr = (uint16_t)(~gfAuto & 0x0000FFFF);	//read as lower 16bits
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
    flashPtr++;
#ifdef MIDI
   *flashPtr = (uint16_t)(~gChannel & 0x0000FFFF);	//read as upper 16bits
#else
   *flashPtr = (uint16_t)(~gfFreq & 0x0000FFFF);	//read as upper 16bits
#endif
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
    flashPtr++;

	k = (uint32_t) gfPWM1 ;
	*flashPtr = (uint16_t)(~k & 0x0000FFFF); //program first=read lower16bit
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	flashPtr ++;
#ifndef MOTOR
	k = (uint32_t) gfPWM2 ;
#else
	k = 0;
#endif
	*flashPtr = (uint16_t)(~k & 0x0000FFFF); //program last=read upper16bit
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
	flashPtr ++;

//now save headers of user functions:
    sramPtr = (uint32_t *) &(gfUserHeader[0]);//252words=1008bytes +16=1024
    for (k = 0;k < 252;k++)  { //first 63 entries,each entry 16bytes=4words
        *flashPtr =(uint16_t)(~(*sramPtr) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        *flashPtr =(uint16_t)((~(*sramPtr) >> 16) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        sramPtr++;
    }
    FLASH->CR &= ~FLASH_CR_PG;
//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

//erase Page30
    FLASH->CR |= FLASH_CR_PER;
    FLASH->AR = PAGE30;
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    FLASH->CR &= ~FLASH_CR_PER;

    FLASH->CR |= FLASH_CR_PG;
	flashPtr = (uint16_t *)PAGE30;
//first write last part of gfUserHeader into page30   starting 0x8007800
    sramPtr = (uint32_t *) &(gfUserHeader[63]);//32*16bytes=128words=512bytes
    for (k = 252;k < 380;k++)  { //entries 64-95,each entry 16bytes=4words
        *flashPtr =(uint16_t)(~(*sramPtr) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        *flashPtr =(uint16_t)((~(*sramPtr) >> 16) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        sramPtr++;
    }
//flashed up to 0x8007A00
//next first part of gfTCode into page30 (512bytes remaining = 128 words)
    sramPtr = (uint32_t *)gfTCode;
    for (k = 0;k < 128;k++)  {
        *flashPtr =(uint16_t)(~(*sramPtr) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        *flashPtr =(uint16_t)(~(*sramPtr >> 16) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        sramPtr++;
    }
    FLASH->CR &= ~FLASH_CR_PG;
//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

//erase Page31
    FLASH->CR |= FLASH_CR_PER;
    FLASH->AR = PAGE31;
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    FLASH->CR &= ~FLASH_CR_PER;

    FLASH->CR |= FLASH_CR_PG;
//write second half of gfTCode into page31  starting 0x8007C00
    flashPtr = (uint16_t *)PAGE31;
    sramPtr = (uint32_t *)gfTCode + 128;	//start offset= 512bytes
    for (k = 0;k < 0x100;k++)  {			//word read, halfWord burn
        *flashPtr =(uint16_t)(~(*sramPtr) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        *flashPtr =(uint16_t)(~(*sramPtr >> 16) & 0x0000FFFF);
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        flashPtr++;
        sramPtr++;
    }
//flash is 100% filled
//terminate
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
}

//###########################################################################

//USART Support
//*************
#ifdef LQFP32
void initUsart1(void) // ****************************************************
// USART1-TxD is PortB,6
// USART1-RxD is PortA,10
// **************************************************************************
{
uint16_t baudset;

    RCC->APB2ENR |=  RCC_APB2ENR_USART1EN ;  // USART1 clock enable

//set PA10 in alternate function mode
    GPIOA->MODER &= ~GPIO_MODER_MODER10_0 ;
    GPIOA->MODER |= GPIO_MODER_MODER10_1;
//set PB6 in alternate function mode
    GPIOB->MODER &= ~GPIO_MODER_MODER6_0 ;
    GPIOB->MODER |= GPIO_MODER_MODER6_1;
//USART1 is AF1 on PortA
    GPIOA->AFR[1] &= 0xFFFFF0FF;        //PA10 is RX
    GPIOA->AFR[1] |= 0x00000100;
//and AF0 on PortB
    GPIOB->AFR[0] &= 0xF0FFFFFF;        //PB6 is TX
//    GPIOB->AFR[0] |= 0x00000000;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;

   switch(gBaud) {
    case 96:
        baudset = 0x1388;       //9600Baud: 5000 at 48 MHz SysClock
        break;
    case 19:
        baudset = 0x09C4;       //19200Baud: 2500 at 48 MHz SysClock
        break;
    case 38:
        baudset = 0x04E2 ;      //38400Baud: 1250 at 48 MHz SysClock
        break;
    case 57:
        baudset = 0x0341;       //57600Baud: 833.33 at 48 MHz SysClock
        break;
    case 11:
    default:
        baudset = 0x01A1;       //112500Baud: 416.67 at 48 MHz SysClock
        break;
    }
    USART1->BRR = baudset;
	USART1->CR2 = USART_CR2_TXINV;	//1 Stopbit, TX level inverted
    USART1->CR3 = 0;				//default:no Handshake
    USART1->CR1 = USART_CR1_RXNEIE | USART_CR1_TE | USART_CR1_RE;
									//8Bit, no Parity, RX Int
    USART1->CR1 |= USART_CR1_UE;	//enable USART
    NVIC_EnableIRQ(USART1_IRQn);
}


void sendB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    while(!(USART1->ISR&USART_ISR_TXE));
    USART1->TDR = data;
}

#else
void initUsart2(void) // ****************************************************
// USART2-TxD is PortA,2
// USART2-RxD is PortA,3
// **************************************************************************
{
uint32_t baudset;

    RCC->APB1ENR |=  RCC_APB1ENR_USART2EN ;   // USART2 clock enable

	GPIOA->MODER &= ~(GPIO_MODER_MODER2_0 |    // Set pin 2,3 in alternate function mode
	                 GPIO_MODER_MODER3_0 ) ;  // 10=Alternate function mode
	GPIOA->MODER |= (GPIO_MODER_MODER2_1 |    // Set pin 2,3 in alternate function mode
	                 GPIO_MODER_MODER3_1 ) ;  // 10=Alternate function mode
    //GPIOA->OTYPER is push-pull by default
    //GPIOA->OSPEEDR is 2 MHz by default
    //GPIOA->PUPDR is HiZ by default (except PA13,14,15)

//USART2 is AF1
    GPIOA->AFR[0] &= 0xFFFF00FF;
    GPIOA->AFR[0] |= 0x00001100;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;
   switch(gBaud) {
    case 96:
        baudset = 0x1380;       //9600Baud: 312.0 at 48 MHz APB Clock
        break;
    case 19:
        baudset = 0x09C0;       //19200Baud: 156.0 at 48 MHz APB Clock
        break;
    case 31:
        baudset = 0x0600 ;      //MIDI=31250Baud: 96.0 at 48 MHz APB Clock
        break;
    case 38:
        baudset = 0x04E0 ;      //38400Baud: 78.0 at 48 MHz APB Clock
        break;
    case 57:
        baudset = 0x0340;       //57600Baud: 52.0 at 48 MHz APB Clock
        break;
    case 11:
    default:
        baudset = 0x01A0;   //115200Baud: 26.0 at 48 MHz APB Clock
    }
    USART2->BRR = baudset;
#ifdef MINI					//the only difference from "MIDI" version
	USART2->CR2 = 0;		//default:1 Stopbit
#else
	USART2->CR2 = USART_CR2_TXINV; 
								//1 Stopbit, TX level inverted
#endif
    USART2->CR3 = 0;			//default:no Handshake
    USART2->CR1 =USART_CR1_UE | USART_CR1_RXNEIE |USART_CR1_TE | USART_CR1_RE;
								//8Bit, no Parity, RX Int
    NVIC_EnableIRQ(USART2_IRQn);
}


void sendB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    while(!(USART2->ISR&USART_ISR_TXE));
    USART2->TDR = data;
}

#endif


#ifdef MIDI
void sendTB(uint8_t data) // ************************************************
//send one byte via serial terminal port (USB and/or USART)
//0=USB and USART both terminal, 1+3= USB is terminal, 2+4=USART is terminal
// **************************************************************************
{
    if (gSplit == 0)  {
        sendB(data);
        if (guConfigured)  sendUsbByte(data);
    }
	else  if ((gSplit == 2) || (gSplit == 4))sendB(data);
							//USART is terminal, USB is serial or MIDI I/O
    else if (guConfigured)  sendUsbByte(data);
							//USB is terminal(virtual COM),USART is serial I/O
}


void sendIOB(uint8_t data) // ************************************************
//send one byte via serial I/O
// **************************************************************************
{
    if ((gSplit == 1) || (gSplit == 3)) { 		//USB is terminal(virtual COM),
        sendB(data); 							//USART is serial I/O
    }
    else if ((gSplit == 2) && (guConfigured)) { //USART=terminal, USB=serial I/O
        sendUsbByte(data);
    }
#ifdef MIDI
    else if ((gSplit == 4) && (guConfigured))  { //USART=terminal, USB=MIDI I/O
        enCodeMidi(data);
	}
#endif
	else gfErrors |= ERR_CONTEXT;	//gSplit=0
}


void sendNibble(uint8_t data) // ********************************************
//0=USB and USART both terminal, 1= USART is terminal, 2=USB is terminal
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
    if (gSplit == 0)  {
        sendB(asciiNibble);
        if (guConfigured)  sendUsbByte(asciiNibble);
    }
    else if (((gSplit == 1) || (gSplit == 3)) && (guConfigured))  {
						//USB is terminal(virtual COM),USART is serial I/O
        sendUsbByte(asciiNibble);
    }
    else sendB(asciiNibble);
}

#else

void sendTB(uint8_t data) // ************************************************
//send one byte via serial terminal port (USB and/or USART)
//gSplit: 0=USB and USART both terminal, 1= only USB is terminal
// **************************************************************************
{
	if (gSplit == 0) sendB(data);
	if (guConfigured)  sendUsbByte(data);
}


void sendNibble(uint8_t data) // ********************************************
//exclusively via terminal
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
	sendTB(asciiNibble);
}

#endif

void sendString(const uint8_t * str) // *************************************
// **************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendTB(byte);
}


void sendX8(uint32_t data) // ***********************************************
// send bits 0...7 in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0x000000F0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    temp= data&0x0000000F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    sendTB(0x20);
}


void sendNX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble, noLead;
    noLead = 0;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    if (nibble)  {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
     if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
}


void send0X32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format WITH leading 0x.Leading zeroes suppressed
// **************************************************************************
{
//	if (data >= 10)  {		//modified 10Jan24
		sendTB('0');
		sendTB('x');
//	}
	sendNX32(data);
}


void sendS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendNS32(data);
    sendTB(0x20);
}


 void sendLS32(uint32_t data) // ********************************************
//send signed decimal number string. Leading zeroes are suppressed
//leading SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendTB(0x20);
    sendNS32(data);
}


void sendNS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//no leading and trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendTB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendTB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendTB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendTB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendTB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendTB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendTB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendTB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendTB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendTB((uint8_t)quotient + 0x30);
ONES:
    sendTB((uint8_t)data + 0x30) ;
}


uint32_t pullUsartRxBuf(void) // ********************************************
// pull 1 byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

	servSystem();   //serve watchdog etc.
    if (gUSART_RXTAIL!=gUSART_RXHEAD)  {
        recv= (uint32_t)((*gUSART_RXTAIL++ & 0x000000FF)|0x00008000);
        if (gUSART_RXTAIL==gUSART_RXBUFTOP)  gUSART_RXTAIL=gUSART_RXBUF;
#ifdef LQFP32
        GPIOA->BSRR = GPIO_BSRR_BR_8; // clear bit PA8,LED OFF
#else
        GPIOB->BSRR = GPIO_BSRR_BR_8; // clear bit PB8,LED OFF
#endif
        gLEDCOUNT = 4; //ca 300 /milliseconds
        return recv;
    }
    else return 0;
}


#ifdef MIDI
uint32_t pullTermRxBuf(void)  // ********************************************
// receive byte from serial terminal port with priority order USB > RS232
// **************************************************************************
{
uint32_t recv;

    if (gSplit == 0)  {
        if (guConfigured)  {            //ASART and USB are parallel terminal
            recv = pullUsbRxBuf();
            if (recv) return recv;      //USB has priority over USART
        }
        recv = pullUsartRxBuf();
        return recv;
    }

	else if ((gSplit == 2)||(gSplit == 4)) { //USART is terminal
            recv = pullUsartRxBuf();
            return recv;
    }

    else if (guConfigured) {            //gSplit = 1 or 3://USB is terminal
        recv = pullUsbRxBuf();  //if (guConfigured == 0): no terminal available
        return recv;
    }
}


uint32_t pullIORxBuf(void)  // *********************************************
// receive byte from serial terminal port with priority order USB > RS232
// *************************************************************************
{
uint32_t recv;

    switch (gSplit)  {
    case 1:
        recv = pullUsartRxBuf();  //USART is serial I/O (COM)
        return recv;
    case 2:                     //USB is serial I/O as virtual COM
        if (guConfigured)  {
            recv = pullUsbRxBuf();
            return recv;
        }
        else return 0;          //if(guConfigured==0) no serial I/O available

    case 0:                     //if gSplit==0: no serial I/O available
    default:
        return  0;
    }
}


void  MidiScanner(uint32_t byte)  // ****************************************
// accepts only MIDI channel as stored in "gChannel", 
// exception: gChannel = 0xFF, then all MIDI channels are accepted
// **************************************************************************
{
static uint32_t status, data1, msgtype;
static uint32_t progress;

/*
	sendTB(CRR);
	sendTB('S');
    sendX8((uint8_t)byte);
    sendX8((uint8_t)progress);
	sendX8((uint8_t)status);
	sendTB('#');
*/

    if (byte >= 0x80) {					//new status byte
			if ((byte == 0xF0) || (byte == EOX) || (byte == 0xF4) ||
						(byte == 0xF5) || (byte == 0xF9) ||(byte == 0xFD) ) {
//								kill SysEx(F0),EOX is end of SysEx message,
//								0xF4, 0xF5, 0xF9 and 0xFD are not defined
				progress = 0;			//reset parser
				status = 0;				//drop input until next status byte
				msgtype = 0;
				return;
			}
			if ((byte == 0xF6) || (byte >= 0xF8))  { //1 byte msg
				gMIDIBUF[gMIDI_HEAD] = byte << 16;		//data bytes = 0
				gMIDI_HEAD++;
				if (gMIDI_HEAD == MIDI_BUFLEN) gMIDI_HEAD = 0;
				progress = 0;			//reset parser
				status = 0;				//drop input until next status byte
				msgtype = 0;
				return;
			}
// "normal" MIDI Channel Msg:
			if (((byte & 0x0000000F) != (gChannel - 1)) && (gChannel != 0)) {
				progress = 0;			//reset parser
				status = 0;				//drop input until next status byte
				msgtype = 0;
				return;
			}	//else accepted status byte:
			status = byte;
			msgtype = byte & 0xF0;
			progress = 1;
	}
	else  {								//new data byte
		if (progress == 0)  {	//SysEx or byte out of context, reset parser
			status = 0;
			msgtype = 0;
			data1 = 0;
			return;
		}
		if (progress == 1)  {
			data1 = byte;
			if ((msgtype == 0xC0)||(msgtype == 0xD0) || (status == 0xF1)||
									(status == 0xF3)) { //2 byte msgs
				gMIDIBUF[gMIDI_HEAD] = (status << 16) | data1;
				gMIDI_HEAD++;					//bits 0..7 for easier parsing
				if (gMIDI_HEAD == MIDI_BUFLEN) gMIDI_HEAD = 0;
				return;		//keep progress==1 and gStatus for Running State
			}
			progress = 2;
			return;
		}
		if (progress == 2)  {					//any 3 byte messages
			gMIDIBUF[gMIDI_HEAD] = (status << 16) | (data1 << 8) | byte;
			gMIDI_HEAD++;					//bits 0..7 for easier parsing
			if (gMIDI_HEAD == MIDI_BUFLEN) gMIDI_HEAD = 0;
			progress = 1;   //reset progress and keep status for Running State
		}
	}
}


uint32_t pullMidiBuf(void) // ********************************************
// pull one word asynchronously from MIDIBUF
// check first if something received. If yes then set bit 31. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

	servSystem();   //serve watchdog etc.

	if (gMIDI_TAIL != gMIDI_HEAD)  {
		recv= gMIDIBUF[gMIDI_TAIL] | 0x80000000;
		gMIDI_TAIL++;
		if (gMIDI_TAIL== MIDI_BUFLEN)  gMIDI_TAIL= 0;
		return recv;
	}
	else return 0;
}

#else

uint32_t pullTermRxBuf(void)  // ********************************************
// receive byte from serial terminal port with priority order USB > RS232
// **************************************************************************
{
uint32_t recv;

	if (guConfigured)  {            //ASART and USB are parallel terminal
		recv = pullUsbRxBuf();
		if (recv) return recv;      //USB has priority over USART
	}
	if (gSplit == 0)  recv = pullUsartRxBuf();
	return recv;
}

#endif  //MIDI or not

// Interrupt Handlers
//*******************
void SysTick_Handler(void)  // ***********************************************
// ***************************************************************************
{
	if (gSYSCOUNT) gSYSCOUNT--;			//controls servSystem()
	if (gDELAYCOUNT) gDELAYCOUNT--;
	if (gTIMEOUT) gTIMEOUT--;			//to be used in any subroutine
	if (gfMSCOUNT) gfMSCOUNT --;		//used by op_MS,op_DS
	if (gfTIXCOUNT) gfTIXCOUNT --;		//used by op_TIX, op_WAIT
#ifdef MOTOR
	if (gfSTEPCOUNT) gfSTEPCOUNT--;		//stepper motor speed
#endif
	gfRandom ^= gfRandom << 13;				//xorshift32 algorithm/white noise
	gfRandom ^= gfRandom >> 17;
	gfRandom ^= gfRandom << 5;
}

#ifdef LQFP32
void USART1_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t recv;

	if (USART1->ISR & USART_ISR_RXNE)  {
		recv = USART1->RDR & 0x00FF;
#ifdef MIDI
		if ((gSplit != 1) && (gSplit != 3))  {	//USART is terminal
			if (recv == CTRL_B) gfTermCtrlBD = CTRL_B;
			if (recv != CTRL_D)  {
				*gUSART_RXHEAD = (uint8_t)recv;
				gUSART_RXHEAD++;
				if (gUSART_RXHEAD==gUSART_RXBUFTOP)gUSART_RXHEAD=gUSART_RXBUF;
			}
			else  op_DEBUG();
		}
#else 
		if (gSplit== 0)  {					//USART is terminal
			if (recv == CTRL_B) gfTermCtrlBD = CTRL_B;
			if (recv != CTRL_D)  {
				*gUSART_RXHEAD = (uint8_t)recv;
				gUSART_RXHEAD++;
				if (gUSART_RXHEAD==gUSART_RXBUFTOP)gUSART_RXHEAD=gUSART_RXBUF;
			}
			else  op_DEBUG();
		}
#endif
		else  {
			*gUSART_RXHEAD = (uint8_t)recv;
			gUSART_RXHEAD++;
			if (gUSART_RXHEAD==gUSART_RXBUFTOP)gUSART_RXHEAD=gUSART_RXBUF;
		}
	}
}

#else

void USART2_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t recv;

	if (USART2->ISR & USART_ISR_RXNE)  {
		recv = USART2->RDR & 0x00FF;
#ifdef MIDI
		if ((gSplit != 1) && (gSplit != 3))  {	//USART is terminal
			if (recv == CTRL_B) gfTermCtrlBD = CTRL_B;
			if (recv != CTRL_D)  {
				*gUSART_RXHEAD = (uint8_t)recv;
				gUSART_RXHEAD++;
				if (gUSART_RXHEAD==gUSART_RXBUFTOP)gUSART_RXHEAD=gUSART_RXBUF;
			}
			else  op_DEBUG();
		}
#else 
		if (gSplit== 0)  {					//USART is terminal
			if (recv == CTRL_B) gfTermCtrlBD = CTRL_B;
			if (recv != CTRL_D)  {
				*gUSART_RXHEAD = (uint8_t)recv;
				gUSART_RXHEAD++;
				if (gUSART_RXHEAD==gUSART_RXBUFTOP)gUSART_RXHEAD=gUSART_RXBUF;
			}
			else  op_DEBUG();
		}
#endif
		else {									//no special char handling
			*gUSART_RXHEAD = (uint8_t)recv;
			gUSART_RXHEAD++;
			if (gUSART_RXHEAD==gUSART_RXBUFTOP)  gUSART_RXHEAD=gUSART_RXBUF;
		}
	}
}
#endif

// #########################################################################
